
import numpy as np
import matplotlib.pyplot as plt
from types import SimpleNamespace
import importlib 
from matplotlib.gridspec import GridSpec
from mpl_toolkits.axes_grid1 import make_axes_locatable
import sys
import os 
from pathlib import Path
if sys.version_info < (3, 0):
    import ConfigParser
else:
    import configparser as ConfigParser
    
import pyzelda.zelda as zelda
import pyzelda.ztools as ztools
import pyzelda.utils.aperture as aperture


from baldrapp.common import baldr_core as bldr
from baldrapp.common import DM_basis as gen_basis
from baldrapp.common import utilities as util


### we generally deal with Simplename spaces since they are lightweight and meet our requirements  
# nevertheless - these can be converted to a class for easier handling at anytime. All functionality should be the same 
class RecursiveNamespaceToClass:
    def __init__(self, namespace):
        # Iterate through all attributes of the SimpleNamespace
        for key, value in vars(namespace).items():
            # If the attribute is another SimpleNamespace, recursively initialize it as an instance of this class
            if isinstance(value, SimpleNamespace):
                value = RecursiveNamespaceToClass(value)
            # Set the attribute on the instance
            setattr(self, key, value)

    def __repr__(self):
        return f"{self.__class__.__name__}({vars(self)})"
    
    
    
################## INTERFACE WITH pyZELDA
""" 
I (git: courtney-barrer) forked the pyZELDA repository and added a new Baldr sensor class that interfaces with the BALDR app.
BaldrApp can work with or without this sensor. If it is appended to the zwfs_ns object, the pyZelda machinary can be used to calculate the intensity.  
But by default I will use my own machinery. This is useful for testing and debugging.
"""


## The most concise way to interface with the pyZelda object is to use the following function to init from ini configuation file
# that contains the necessary information to configure the ZWFS.
config_ini = '/home/benja/Documents/BALDR/BaldrApp/baldrapp/configurations/baldr_1.ini'
zwfs_ns = bldr.init_zwfs_from_config_ini( config_ini=config_ini , wvl0=1.25e-6)

# pyZelda object is here 
zwfs_ns.pyZelda

# this is the same object as the following (using the instrument specified in config_ini)
# configure the ZWFS for the UT 
z = zelda.Sensor('BALDR_UT_J3')


# IF YOU DO THINGS MANUALLY (not using .ini file - not recommended): 
# Need to init things consistently with the pyZelda object.
# 
# our own initialization of the zwfs using simple namespace due to speed
wvl0 = 1.25e-6
grid_ns, optics_ns = bldr.init_ns_from_pyZelda(z, wvl0)

dm_dict = {
    "dm_model":"BMC-multi-3.5",
    "actuator_coupling_factor":1.1, #0.7,# std of in actuator spacing of gaussian IF applied to each actuator. (e.g actuator_coupling_factor = 1 implies std of poke is 1 actuator across.)
    "dm_pitch":1,
    "dm_aoi":0, # angle of incidence of light on DM 
    "opd_per_cmd" : 3e-6, # peak opd applied at center of actuator per command unit (normalized between 0-1) 
    "flat_rmse" : 20e-9 # std (m) of flatness across Flat DM  
    }


dm_ns = SimpleNamespace(**dm_dict)

zwfs_ns = bldr.init_zwfs(grid_ns, optics_ns, dm_ns)

# check if consistent:
consistency_log = bldr.check_ns_consistency_with_pyZelda( z, zwfs_ns )
if consistency_log=={}:
    print('ALL GOOD!')

# we can simply append the pyZelda object z to our zwfs_ns namespace. This can be converted to a class simply 
zwfs_ns.pyZelda = z

# if we want (it is not necessary) we can convert the namespace to a class
zwfs_class = RecursiveNamespaceToClass( zwfs_ns )

# all functionality should remain the same 
# e.g. 
zwfs_class.pyZelda.mask_phase_shift(1.25e-6)

plt.figure(1)
plt.imshow( zwfs_ns.grid.pupil_mask )
plt.figure(2)
plt.imshow( zwfs_class.pyZelda.pupil)


phi, phi_internal,  N0, I0, Intensity = bldr.test_propagation( zwfs_ns )

pupil_roi = z.pupil
z_opd_advanced = z.analyze(clear_pupil=N0, zelda_pupil = Intensity , wave=zwfs_ns.optics.wvl0,
                           use_arbitrary_amplitude=True,
                           refwave_from_clear=True,
                           cpix=False, pupil_roi=pupil_roi)


opd_input = 40e-9* zwfs_ns.grid.pupil_mask *  np.random.randn( *zwfs_ns.grid.pupil_mask.shape) #nm


import time
start_time = time.time()
I = z.propagate_opd_map( 1 * opd_input, wave=zwfs_ns.optics.wvl0 )
end_time = time.time()
print( end_time - start_time )
plt.imshow( I  ) ; plt.show()


start_time = time.time()
I2 = bldr.get_frame(  opd_input  = opd_input ,   amp_input = zwfs_ns.grid.pupil_mask ,\
    opd_internal = 0* opd_input,  zwfs_ns= zwfs_ns , detector=None )
end_time = time.time()
print( end_time - start_time )
plt.imshow( I2  ) ; plt.show()

# slower because do DM surface processing when getting frame

b, expi = ztools.create_reference_wave_beyond_pupil(z.mask_diameter, z.mask_depth, z.mask_substrate, z.mask_Fratio,
                                       z.pupil_diameter, z.pupil, zwfs_ns.optics.wvl0, clear=np.array([]), 
                                       sign_mask=np.array([]), cpix=False)



b_ab, expi = ztools.create_reference_wave_beyond_pupil_with_aberrations( opd_input, z.mask_diameter, z.mask_depth, z.mask_substrate, z.mask_Fratio,
                                       z.pupil_diameter, z.pupil, zwfs_ns.optics.wvl0, clear=np.array([]), 
                                       sign_mask=np.array([]), cpix=False)

plt.imshow (np.abs( b )); plt.colorbar(); plt.show()

plt.imshow (np.abs( b_ab )); plt.colorbar(); plt.show()



plt.figure(1)
plt.imshow( N0 )

plt.figure(2)
plt.imshow( I0 )

plt.figure(3)
plt.imshow( z_opd_advanced )
plt.show()


################## TEST 1
# check dm registration on pupil (wavespace)
zwfs_ns = bldr.init_zwfs(grid_ns, optics_ns, dm_ns)

    
################## TEST 0 
# configure our zwfs 
grid_dict = {
    "telescope":'UT',
    "D":1, # diameter of beam 
    "N" : 64, # number of pixels across pupil diameter
    "dim": 64 * 4
    #"padding_factor" : 4, # how many pupil diameters fit into grid x axis
    # TOTAL NUMBER OF PIXELS = padding_factor * N 
    }

optics_dict = {
    "wvl0" :1.65e-6, # central wavelength (m) 
    "F_number": 21.2, # F number on phasemask
    "mask_diam": 1.06, # diameter of phaseshifting region in diffraction limit units (physical unit is mask_diam * 1.22 * F_number * lambda)
    "theta": 1.57079, # phaseshift of phasemask 
}

dm_dict = {
    "dm_model":"BMC-multi-3.5",
    "actuator_coupling_factor":0.7,# std of in actuator spacing of gaussian IF applied to each actuator. (e.g actuator_coupling_factor = 1 implies std of poke is 1 actuator across.)
    "dm_pitch":1,
    "dm_aoi":0, # angle of incidence of light on DM 
    "opd_per_cmd" : 3e-6, # peak opd applied at center of actuator per command unit (normalized between 0-1) 
    "flat_rmse" : 20e-9 # std (m) of flatness across Flat DM  
    }

grid_ns = SimpleNamespace(**grid_dict)
optics_ns = SimpleNamespace(**optics_dict)
dm_ns = SimpleNamespace(**dm_dict)




################## TEST 1
# check dm registration on pupil (wavespace)

#zwfs_ns = bldr.init_zwfs(grid_ns, optics_ns, dm_ns) # to do it manually on the most basic level    
zwfs_ns = bldr.init_zwfs_from_config_ini( config_ini=config_ini , wvl0=1.25e-6)

zwfs_class = RecursiveNamespaceToClass( zwfs_ns )

phi, phi_internal,  N0, I0, Intensity = bldr.test_propagation( zwfs_ns )

#phi, phi_internal,  N0, I0, Intensity = bldr.test_propagation( zwfs_class )

fig = plt.figure() 
im = plt.imshow( N0, extent=[np.min(zwfs_ns.grid.wave_coord.x), np.max(zwfs_ns.grid.wave_coord.x),\
    np.min(zwfs_ns.grid.wave_coord.y), np.max(zwfs_ns.grid.wave_coord.y)] )
cbar = fig.colorbar(im, ax=plt.gca(), pad=0.01)
cbar.set_label(r'Pupil Intensity', fontsize=15, labelpad=10)

plt.scatter(zwfs_ns.grid.dm_coord.act_x0_list_wavesp, zwfs_ns.grid.dm_coord.act_y0_list_wavesp, color='blue', marker='.', label = 'DM actuators')
plt.show() 

################## TEST 2
# check pupil intensities (not using pyZelda machinery)

fig,ax = plt.subplots( 1,4 )
ax[0].imshow( util.get_DM_command_in_2D( zwfs_ns.dm.current_cmd ))
ax[0].set_title('dm cmd')
ax[1].set_title('OPD wavespace')
ax[1].imshow( phi )
ax[2].set_title('ZWFS Intensity')
ax[2].imshow( Intensity )
ax[3].set_title('ZWFS reference Intensity')
ax[3].imshow( I0 )


################## TEST 3 
# test updating the DM registration 
#zwfs_ns = bldr.init_zwfs(grid_ns, optics_ns, dm_ns)
zwfs_ns = bldr.init_zwfs_from_config_ini( config_ini=config_ini , wvl0=1.25e-6)

# redefining the affine transform between DM coordinates and the wavefront space
#a, b, c, d = 1.8*zwfs_ns.grid.D/np.ptp(zwfs_ns.grid.wave_coord.x), 0, 0, 1.8*grid_ns.D/np.ptp(zwfs_ns.grid.wave_coord.x)  # Parameters for affine transform (identity for simplicity)

a, b, c, d = 1.8*zwfs_ns.grid.D/np.ptp(zwfs_ns.grid.wave_coord.x), 0, 0, 1.8*zwfs_ns.grid.D/np.ptp(zwfs_ns.grid.wave_coord.x)  # Parameters for affine transform (identity for simplicity)

# offset 5% of pupil 
t_x, t_y = np.mean(zwfs_ns.grid.wave_coord.x) + 0.05 * zwfs_ns.grid.D, np.mean(zwfs_ns.grid.wave_coord.x)  # Translation in phase space

# we could also introduce mis-registrations by rolling input pupil 
dm_act_2_wave_space_transform_matrix = np.array( [[a,b,t_x],[c,d,t_y]] )

zwfs_ns = bldr.update_dm_registration_wavespace( dm_act_2_wave_space_transform_matrix , zwfs_ns )

phi, phi_internal,  N0, I0, Intensity = bldr.test_propagation( zwfs_ns )


fig = plt.figure() 
im = plt.imshow( N0, extent=[np.min(zwfs_ns.grid.wave_coord.x), np.max(zwfs_ns.grid.wave_coord.x),\
    np.min(zwfs_ns.grid.wave_coord.y), np.max(zwfs_ns.grid.wave_coord.y)] )
cbar = fig.colorbar(im, ax=plt.gca(), pad=0.01)
cbar.set_label(r'Pupil Intensity', fontsize=15, labelpad=10)

plt.scatter(zwfs_ns.grid.dm_coord.act_x0_list_wavesp, zwfs_ns.grid.dm_coord.act_y0_list_wavesp, color='blue', marker='.', label = 'DM actuators')
plt.show() 



################## TEST 4
# test DM basis generation 
basis_name_list = ['Hadamard', "Zonal", "Zonal_pinned_edges", "Zernike", "Zernike_pinned_edges", "fourier", "fourier_pinned_edges"]

fig, ax = plt.subplots( len( basis_name_list), len( basis_name_list) ,figsize=(15,15) )
for i,b in enumerate( basis_name_list ) :
    
    print( b )
    basis_test = gen_basis.construct_command_basis( basis= b, number_of_modes = 20, Nx_act_DM = 12, Nx_act_basis = 12, act_offset=(0,0), without_piston=True)
    
    for m in range( ax.shape[1] ):
        ax[ i, m ].imshow( util.get_DM_command_in_2D( basis_test.T[m] ) )
    
    ax[i, 0].set_ylabel( b )
    #print( basis_test )
    
#plt.savefig( "/Users/bencb/Downloads/baldr_bases.png",dpi=300)
plt.show() 



################## TEST 5
# Get reference intensities 
#zwfs_ns = bldr.init_zwfs(grid_ns, optics_ns, dm_ns)
zwfs_ns = bldr.init_zwfs_from_config_ini( config_ini=config_ini , wvl0=1.25e-6)

opd_input = 0 * zwfs_ns.grid.pupil_mask *  np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

opd_internal = 10e-9 * zwfs_ns.grid.pupil_mask * np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

amp_input = 1e2 * zwfs_ns.grid.pupil_mask

I0 = bldr.get_I0(  opd_input  = opd_input ,   amp_input = amp_input,\
    opd_internal = opd_internal,  zwfs_ns= zwfs_ns , detector=None )

N0 = bldr.get_N0(  opd_input  = opd_input ,   amp_input = amp_input,\
    opd_internal = opd_internal,  zwfs_ns= zwfs_ns , detector=None )

# converting to a class and running the same function 
zwfs_class = RecursiveNamespaceToClass( zwfs_ns )

I0 = bldr.get_I0(  opd_input  = opd_input ,   amp_input = amp_input,\
    opd_internal = opd_internal,  zwfs_ns= zwfs_class , detector=None )

N0 = bldr.get_N0(  opd_input  = opd_input ,   amp_input = amp_input,\
    opd_internal = opd_internal,  zwfs_ns= zwfs_class , detector=None )

plt.figure(); plt.imshow( I0 ) ;plt.show()

################## TEST 6
# classify pupil regions and plot them
#zwfs_ns = bldr.init_zwfs(grid_ns, optics_ns, dm_ns)
zwfs_ns = bldr.init_zwfs_from_config_ini( config_ini=config_ini , wvl0=1.25e-6)

opd_input = 0 * zwfs_ns.grid.pupil_mask *  np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

opd_internal = 10e-9 * zwfs_ns.grid.pupil_mask * np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

amp_input = 1e2 * zwfs_ns.grid.pupil_mask

zwfs_ns = bldr.classify_pupil_regions( opd_input,  amp_input ,  opd_internal,  zwfs_ns , detector=None)

fig,ax = plt.subplots( 1, len(zwfs_ns.pupil_regions.__dict__ ) ,figsize=(10,10) )
for k,axx in zip( zwfs_ns.pupil_regions.__dict__,ax.reshape(-1)):
    axx.imshow(zwfs_ns.pupil_regions.__dict__[k] )
    axx.set_title( k ) 
plt.show()


################## TEST 7
# Build IM  and look at Eigenmodes! 
#zwfs_ns = bldr.init_zwfs(grid_ns, optics_ns, dm_ns)
zwfs_ns = bldr.init_zwfs_from_config_ini( config_ini=config_ini , wvl0=1.25e-6)

# converting to a class and running the same function 
zwfs_class = RecursiveNamespaceToClass( zwfs_ns )

opd_input = 0 * zwfs_ns.grid.pupil_mask *  np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

opd_internal = 10e-9 * zwfs_ns.grid.pupil_mask * np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

amp_input = 1e4 * zwfs_ns.grid.pupil_mask

# we must first define our pupil regions before building 
zwfs_ns = bldr.classify_pupil_regions( opd_input,  amp_input ,  opd_internal,  zwfs_ns , detector=None)

# using  class
#_ = bldr.classify_pupil_regions( opd_input,  amp_input ,  opd_internal,  zwfs_class , detector=None)

basis_name_list = ['Hadamard', "Zonal", "Zonal_pinned_edges", "Zernike", "Zernike_pinned_edges", "fourier", "fourier_pinned_edges"]

# perfect field only with internal opd aberrations 
# different poke methods 
Nmodes = 100
basis = 'Zonal_pinned_edges'
M2C_0 = gen_basis.construct_command_basis( basis= basis, number_of_modes = Nmodes, without_piston=True).T  


_ = bldr.build_IM( zwfs_ns ,  calibration_opd_input= 0 *zwfs_ns.grid.pupil_mask , calibration_amp_input = amp_input , \
    opd_internal = opd_internal,  basis = basis, Nmodes =  Nmodes, poke_amp = 0.05, poke_method = 'double_sided_poke',\
        imgs_to_mean = 1, detector=None)

_ = bldr.build_IM( zwfs_ns ,  calibration_opd_input= 0 *zwfs_ns.grid.pupil_mask , calibration_amp_input = amp_input , \
    opd_internal = opd_internal,  basis = basis, Nmodes =  Nmodes, poke_amp = 0.05, poke_method = 'single_sided_poke',\
        imgs_to_mean = 1, detector=None)

# different basis 
basis = 'Hadamard'
M2C_0 = gen_basis.construct_command_basis( basis= basis, number_of_modes = Nmodes, without_piston=True).T  

_ = bldr.build_IM( zwfs_ns ,  calibration_opd_input= 0 *zwfs_ns.grid.pupil_mask , calibration_amp_input = amp_input , \
    opd_internal = opd_internal,  basis = basis, Nmodes =  Nmodes, poke_amp = 0.05, poke_method = 'double_sided_poke',\
        imgs_to_mean = 1, detector=None)


# build IM with zwfs_class
_ = bldr.build_IM( zwfs_class ,  calibration_opd_input= 0 *zwfs_ns.grid.pupil_mask , calibration_amp_input = amp_input , \
    opd_internal = opd_internal,  basis = basis, Nmodes =  Nmodes, poke_amp = 0.05, poke_method = 'double_sided_poke',\
        imgs_to_mean = 1, detector=None)


################## TEST 7.5
# look at reference field (b)
# Build IM  and look at Eigenmodes! 
zwfs_ns = bldr.init_zwfs(grid_ns, optics_ns, dm_ns)

# converting to a class and running the same function 
zwfs_class = RecursiveNamespaceToClass( zwfs_ns )

opd_input = 1 * zwfs_ns.grid.pupil_mask *  np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

opd_internal = 10e-9 * zwfs_ns.grid.pupil_mask * np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

amp_input = 1e4 * zwfs_ns.grid.pupil_mask

pupil_basis = ztools.zernike.zernike_basis(nterms=15, npix=z.pupil_diameter, rho=None, theta=None) * 1e-9

opd_input = 500 * z.pupil * util.insert_concentric(  np.nan_to_num( pupil_basis[10] ) , np.zeros( z.pupil.shape) )

# something wrong here 
psi_b = bldr.get_psf( phi= 2*3.14 * opd_input /zwfs_ns.optics.wvl0, phasemask_diameter = zwfs_ns.optics.mask_diam, pupil_diameter = zwfs_ns.grid.N, \
         fplane_pixels=zwfs_ns.focal_plane.fplane_pixels, pixels_across_mask=zwfs_ns.focal_plane.pixels_across_mask )
plt.imshow( abs( psi_b )**2, extent=[-1, 1, -1, 1], origin='lower' ); plt.colorbar(); plt.show()

b =  bldr.get_b( phi= opd_input, phasemask= zwfs_ns.grid.phasemask_mask, phasemask_diameter = zwfs_ns.optics.mask_diam, pupil_diameter = zwfs_ns.grid.N, \
         fplane_pixels=zwfs_ns.focal_plane.fplane_pixels, pixels_across_mask=zwfs_ns.focal_plane.pixels_across_mask )
plt.imshow( abs( b ), extent=[-1, 1, -1, 1], origin='lower' ); plt.colorbar(); plt.show()



psi_A = np.exp( 1j *  2*3.14 * opd_input /zwfs_ns.optics.wvl0)

plt.imshow( np.fft.fftshift( abs( np.fft.fft2( np.pad( psi_A, psi_A.shape ) ) ) )); plt.colorbar(); plt.show()


################## TEST 8
# project onto TT and HO

# Build IM  and look at Eigenmodes! 
zwfs_ns = bldr.init_zwfs(grid_ns, optics_ns, dm_ns)

opd_input = 0 * zwfs_ns.grid.pupil_mask *  np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

opd_internal = 10e-9 * zwfs_ns.grid.pupil_mask * np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

amp_input = 1e4 * zwfs_ns.grid.pupil_mask


#basis_name_list = ['Hadamard', "Zonal", "Zonal_pinned_edges", "Zernike", "Zernike_pinned_edges", "fourier", "fourier_pinned_edges"]

# perfect field only with internal opd aberrations 
# different poke methods 
Nmodes = 100
basis = 'Zonal_pinned_edges'
poke_amp = 0.05
Smax = 30
detector = (4,4) # for binning , zwfs_ns.grid.N is #pixels across pupil diameter (64) therefore division 4 = 16 pixels (between CRed2 and Cred1 )
#M2C_0 = gen_basis.construct_command_basis( basis= basis, number_of_modes = Nmodes, without_piston=True).T  

#I0 = bldr.get_I0(  opd_input  = 0 *zwfs_ns.grid.pupil_mask ,   amp_input = amp_input,\
#    opd_internal = opd_internal,  zwfs_ns= zwfs_ns , detector=None )

#N0 = bldr.get_N0(  opd_input  = 0 *zwfs_ns.grid.pupil_mask  ,   amp_input = amp_input,\
#    opd_internal = opd_internal,  zwfs_ns= zwfs_ns , detector=None )

# we must first define our pupil regions before building 
zwfs_ns = bldr.classify_pupil_regions( opd_input,  amp_input ,  opd_internal,  zwfs_ns , detector= detector) # For now detector is just tuple of pixels to average. useful to know is zwfs_ns.grid.N is number of pixels across pupil. # from this calculate an appropiate binning for detector 

zwfs_ns = bldr.build_IM( zwfs_ns,  calibration_opd_input= 0 *zwfs_ns.grid.pupil_mask , calibration_amp_input = amp_input , \
    opd_internal = opd_internal,  basis = basis, Nmodes =  Nmodes, poke_amp = poke_amp, poke_method = 'double_sided_poke',\
        imgs_to_mean = 1, detector=detector )

# look at the eigenmodes in camera, DM and singular values
bldr.plot_eigenmodes( zwfs_ns , save_path = None )

TT_vectors = gen_basis.get_tip_tilt_vectors()

#zwfs_ns = bldr.construct_ctrl_matricies_from_IM(zwfs_ns,  method = 'Eigen_TT-HO', Smax = 50, TT_vectors = TT_vectors )
zwfs_ns = bldr.construct_ctrl_matricies_from_IM(zwfs_ns,  method = 'Eigen_TT-HO', Smax = 20, TT_vectors = TT_vectors )

#zwfs_ns = bldr.add_controllers_for_MVM_TT_HO( zwfs_ns, TT = 'PID', HO = 'leaky')
zwfs_ns = bldr.add_controllers_for_MVM_TT_HO( zwfs_ns, TT = 'PID', HO = 'leaky')

#zwfs_ns = init_CL_simulation( zwfs_ns,  opd_internal, amp_input , basis, Nmodes, poke_amp, Smax )
            
dm_disturbance = 0.1 * TT_vectors.T[0]
#zwfs_ns.dm.current_cmd =  zwfs_ns.dm.dm_flat + disturbance_cmd 

# as example how to reset telemetry
zwfs_ns.dm.current_cmd = zwfs_ns.dm.dm_flat + dm_disturbance
zwfs_ns = bldr.reset_telemetry( zwfs_ns )
zwfs_ns.ctrl.TT_ctrl.reset()
zwfs_ns.ctrl.HO_ctrl.reset()
zwfs_ns.ctrl.TT_ctrl.set_all_gains_to_zero()
zwfs_ns.ctrl.HO_ctrl.set_all_gains_to_zero()

close_after = 20
kwargs = {"I0":zwfs_ns.reco.I0, "HO_ctrl": zwfs_ns.ctrl.HO_ctrl, "TT_ctrl": zwfs_ns.ctrl.TT_ctrl }
for i in range(100):
    print(f'iteration {i}')
    if i > close_after : 
        #zwfs_ns.ctrl.HO_ctrl.ki = 0.2 * np.ones( len(zwfs_ns.ctrl.HO_ctrl.ki) )
        #zwfs_ns.ctrl.HO_ctrl.kp = 1 * np.ones( len(zwfs_ns.ctrl.HO_ctrl.kp) )

        zwfs_ns.ctrl.TT_ctrl.kp = 1 * np.ones( len(zwfs_ns.ctrl.TT_ctrl.kp) )
        zwfs_ns.ctrl.TT_ctrl.ki = 0.8 * np.ones( len(zwfs_ns.ctrl.TT_ctrl.ki) )
        
    
    bldr.AO_iteration( opd_input, amp_input, opd_internal,  zwfs_ns, dm_disturbance,  record_telemetry=True , method = 'MVM-TT-HO', detector=detector, **kwargs)

# Generate some data


i = len(zwfs_ns.telem.rmse_list) - 1
plt.ioff() 
        
#for i in range(10):
im_dm_dist = util.get_DM_command_in_2D( zwfs_ns.telem.dm_disturb_list[i] )
im_phase = zwfs_ns.telem.field_phase[i]
im_int = zwfs_ns.telem.i_list[i]
im_cmd = util.get_DM_command_in_2D( np.array(zwfs_ns.telem.c_TT_list[i]) + np.array(zwfs_ns.telem.c_HO_list[i])  ) 


#line_x = np.linspace(0, i, i)
line_eHO = zwfs_ns.telem.e_HO_list[:i]
line_eTT = zwfs_ns.telem.e_TT_list[:i]
line_S = zwfs_ns.telem.strehl[:i]
line_rmse = zwfs_ns.telem.rmse_list[:i]

# Define plot data
image_list = [im_dm_dist, im_phase, im_int, im_cmd]
image_title_list = ['DM disturbance', 'input phase', 'intensity', 'reco. command']
image_colorbar_list = ['DM units', 'radians', 'adu', 'DM units']

plot_list = [ line_eHO, line_eTT, line_S, line_rmse ] 
plot_ylabel_list = ['e_HO', 'e_TT', 'Strehl', 'rmse']
plot_xlabel_list = ['iteration' for _ in plot_list]
plot_title_list = ['' for _ in plot_list]

#vlims = [(0, 1), (0, 1), (0, 1)]  # Set vmin and vmax for each image

util.create_telem_mosaic(image_list, image_title_list, image_colorbar_list, 
                plot_list, plot_title_list, plot_xlabel_list, plot_ylabel_list)







#####################
## REAL TIME SIMULATION APP 
####################


import sys
from PyQt5 import QtWidgets, QtCore
import pyqtgraph as pg


class AOControlApp(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()

        # Layout
        self.layout = QtWidgets.QVBoxLayout()

        # Buttons
        self.run_button = QtWidgets.QPushButton("Run")
        self.pause_button = QtWidgets.QPushButton("Pause")
        self.zero_gains_button = QtWidgets.QPushButton("Set Gains to Zero")
        self.reset_button = QtWidgets.QPushButton("reset")
        
        # Connect buttons to functions
        self.run_button.clicked.connect(self.run_loop)
        self.pause_button.clicked.connect(self.pause_loop)
        self.zero_gains_button.clicked.connect(self.set_gains_to_zero)
        self.reset_button.clicked.connect(self.reset)

        # Text input for user
        self.input_label = QtWidgets.QLabel("User Input:")
        self.text_input = QtWidgets.QLineEdit()
        self.text_input.returnPressed.connect(self.check_input)

        # Add buttons and input to layout
        self.layout.addWidget(self.run_button)
        self.layout.addWidget(self.pause_button)
        self.layout.addWidget(self.zero_gains_button)
        self.layout.addWidget(self.reset_button)
        self.layout.addWidget(self.input_label)
        self.layout.addWidget(self.text_input)

        # PyQtGraph plots
        self.plot_widget = pg.GraphicsLayoutWidget()
        self.layout.addWidget(self.plot_widget)

        # Create Image and Line plots in PyQtGraph
        self.image_plots = []
        for i in range(4):
            img_view = self.plot_widget.addPlot(row=0, col=i)
            img_item = pg.ImageItem()
            img_view.addItem(img_item)
            self.image_plots.append(img_item)

        self.line_plots = []
        for i in range(4):
            line_plot = self.plot_widget.addPlot(row=1 + (i // 2), col=(i % 2) * 2, colspan=2)
            self.line_plots.append(line_plot.plot())

        # Add the layout to the widget
        self.setLayout(self.layout)

        # Timer for running the AO_iteration in a loop
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.run_AO_iteration)
        self.loop_running = False

    def run_loop(self):
        if not self.loop_running:
            self.loop_running = True
            self.timer.start(100)  # Adjust time in ms if needed

    def pause_loop(self):
        self.loop_running = False
        self.timer.stop()

    def set_gains_to_zero(self):
        self.pause_loop()
        zwfs_ns.ctrl.TT_ctrl.set_all_gains_to_zero()
        zwfs_ns.ctrl.HO_ctrl.set_all_gains_to_zero()
        self.run_loop()

    def reset(self):
        self.pause_loop()
        zwfs_ns = bldr.reset_telemetry( zwfs_ns )
        zwfs_ns.ctrl.TT_ctrl.reset()
        zwfs_ns.ctrl.HO_ctrl.reset()
        
    def check_input(self):
        user_input = self.text_input.text()
        self.pause_loop()
        # Placeholder: Add conditions for user input processing
        print(f"User input: {user_input}")  # Replace with actual processing logic
        
        if 'kpHO*=' in user_input:
            factor = float( user_input.split('=')[-1] )
            zwfs_ns.ctrl.HO_ctrl.kp *= factor
        if 'kpTT*=' in user_input:
            factor = float( user_input.split('=')[-1] )
            zwfs_ns.ctrl.TT_ctrl.kp *= factor
        if 'kiHO*=' in user_input:
            factor = float( user_input.split('=')[-1] )
            zwfs_ns.ctrl.HO_ctrl.ki *= factor
        if 'kiTT*=' in user_input:
            factor = float( user_input.split('=')[-1] )
            zwfs_ns.ctrl.TT_ctrl.ki *= factor

        
        if 'kpHO[' in user_input:
            index = int( user_input.split('[')[1].split(']')[0] )
            value = float( user_input.split('=')[-1] )
            zwfs_ns.ctrl.HO_ctrl.kp[index] = value
        if 'kpTT[' in user_input:
            index = int( user_input.split('[')[1].split(']')[0] )
            value = float( user_input.split('=')[-1] )
            zwfs_ns.ctrl.TT_ctrl.kp[index] = value
        if 'kiHO[' in user_input:
            index = int( user_input.split('[')[1].split(']')[0] )
            value = float( user_input.split('=')[-1] )
            zwfs_ns.ctrl.HO_ctrl.ki[index] = value
        if 'kiTT[' in user_input:
            index = int( user_input.split('[')[1].split(']')[0] )
            value = float( user_input.split('=')[-1] )
            zwfs_ns.ctrl.TT_ctrl.ki[index] = value

                    
        self.run_loop()

    def run_AO_iteration(self):
        
        kwargs = {"I0":zwfs_ns.reco.I0, "HO_ctrl": zwfs_ns.ctrl.HO_ctrl, "TT_ctrl": zwfs_ns.ctrl.TT_ctrl }
        # Call the AO iteration function from your module
        bldr.AO_iteration( opd_input, amp_input, opd_internal,  zwfs_ns, dm_disturbance, record_telemetry=True , method = 'MVM-TT-HO', detector=detector, **kwargs)


        # Retrieve telemetry data
        im_dm_dist = util.get_DM_command_in_2D(zwfs_ns.telem.dm_disturb_list[-1])
        im_phase = zwfs_ns.telem.field_phase[-1]
        im_int = zwfs_ns.telem.i_list[-1]
        im_cmd = util.get_DM_command_in_2D(np.array(zwfs_ns.telem.c_TT_list[-1]) + np.array(zwfs_ns.telem.c_HO_list[-1]))

        # Update images in the PyQtGraph interface
        self.image_plots[0].setImage(im_dm_dist)
        self.image_plots[1].setImage(im_phase)
        self.image_plots[2].setImage(im_int)
        self.image_plots[3].setImage(im_cmd)

        # Update line plots
        # Check if line data exists
        if len(zwfs_ns.telem.e_HO_list) > 0:
            self.line_plots[0].setData(zwfs_ns.telem.e_HO_list)
            self.line_plots[0].getViewBox().autoRange()  # Force autoscaling of the plot

        if len(zwfs_ns.telem.e_TT_list) > 0:
            self.line_plots[1].setData(zwfs_ns.telem.e_TT_list)
            self.line_plots[1].getViewBox().autoRange()

        if len(zwfs_ns.telem.strehl) > 0:
            self.line_plots[2].setData(zwfs_ns.telem.strehl)
            self.line_plots[2].getViewBox().autoRange()

        if len(zwfs_ns.telem.rmse_list) > 0:
            self.line_plots[3].setData(zwfs_ns.telem.rmse_list)
            self.line_plots[3].getViewBox().autoRange()


if __name__ == "__main__":
    
    zwfs_ns = bldr.reset_telemetry( zwfs_ns )
    zwfs_ns.ctrl.TT_ctrl.reset()
    zwfs_ns.ctrl.HO_ctrl.reset()
    zwfs_ns.ctrl.TT_ctrl.set_all_gains_to_zero()
    zwfs_ns.ctrl.HO_ctrl.set_all_gains_to_zero()

    zwfs_ns.ctrl.HO_ctrl.ki = 0.2 * np.ones( len(zwfs_ns.ctrl.HO_ctrl.ki) )
    zwfs_ns.ctrl.HO_ctrl.kp = 1 * np.ones( len(zwfs_ns.ctrl.HO_ctrl.kp) )

    zwfs_ns.ctrl.TT_ctrl.kp = 1 * np.ones( len(zwfs_ns.ctrl.TT_ctrl.kp) )
    zwfs_ns.ctrl.TT_ctrl.ki = 0.8 * np.ones( len(zwfs_ns.ctrl.TT_ctrl.ki) )
    


    app = QtWidgets.QApplication(sys.argv)
    window = AOControlApp()
    window.setWindowTitle("AO Control GUI")
    window.show()
    sys.exit(app.exec_())












"""
import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtWidgets
import numpy as np
import sys

# Data placeholders for demonstration
dm_disturb_list = [np.random.randn(12, 12) for _ in range(100)]
i_list = [np.random.rand(64, 64) for _ in range(100)]
e_TT_list = [np.random.randn(10) for _ in range(100)]
e_HO_list = [np.random.randn(20) for _ in range(100)]
rmse_list = np.random.rand(100)

# Create a Qt Application
app = QtWidgets.QApplication([])

# Create the main window and layout
main_win = QtWidgets.QWidget()
main_layout = QtWidgets.QVBoxLayout()
main_win.setLayout(main_layout)

# Create the pyqtgraph GraphicsLayoutWidget for plots
plot_win = pg.GraphicsLayoutWidget(show=True, title="Real-Time Plotting with PyQtGraph")
plot_win.resize(1000, 800)
plot_win.setWindowTitle('Real-Time ZWFS Control Loops')

main_layout.addWidget(plot_win)

# Create the image plot for DM disturbance
dm_plot = plot_win.addPlot(title="DM Disturbance")
dm_img = pg.ImageItem()
dm_plot.addItem(dm_img)

# Create the image plot for i_list (camera image)
i_list_plot = plot_win.addPlot(title="Camera Image")
i_img = pg.ImageItem()
i_list_plot.addItem(i_img)
plot_win.nextRow()

# Create the line plot for e_TT signal
e_TT_plot = plot_win.addPlot(title="Tip/Tilt Error Signal (e_TT)")
e_TT_curve = e_TT_plot.plot()

# Create the line plot for e_HO signal
e_HO_plot = plot_win.addPlot(title="Higher Order Error Signal (e_HO)")
e_HO_curve = e_HO_plot.plot()
plot_win.nextRow()

# Create the line plot for RMSE
rmse_plot = plot_win.addPlot(title="RMSE")
rmse_curve = rmse_plot.plot()

# Now create a horizontal layout for buttons
button_layout = QtWidgets.QHBoxLayout()

# Run Button
run_button = QtWidgets.QPushButton('Run')
button_layout.addWidget(run_button)

# Pause Button
pause_button = QtWidgets.QPushButton('Pause')
button_layout.addWidget(pause_button)

# Reset Button
reset_button = QtWidgets.QPushButton('Reset')
button_layout.addWidget(reset_button)

# Add the button panel to the main layout
main_layout.addLayout(button_layout)

# Initialize some settings
idx = 0
paused = False  # To handle pause

def update_plots(i):
    # Update DM disturbance image
    dm_img.setImage(dm_disturb_list[i])

    # Update i_list (camera image)
    i_img.setImage(i_list[i])

    # Update e_TT signal (1D)
    e_TT_curve.setData(e_TT_list[i])

    # Update e_HO signal (1D)
    e_HO_curve.setData(e_HO_list[i])

    # Update RMSE curve
    rmse_curve.setData(rmse_list[:i+1])

# Timer to update the plots in real-time
def update():
    global idx, paused
    if paused:
        return  # Do nothing if paused
    if idx >= len(i_list):  # Stop when finished
        return
    update_plots(idx)
    idx += 1

# Run button event handler
def run():
    global paused
    paused = False  # Resume the update
    timer.start(100)

# Pause button event handler
def pause():
    global paused
    paused = True  # Stop the update

# Reset button event handler
def reset():
    global idx, paused
    idx = 0  # Reset index to start over
    paused = True  # Stop the update
    update_plots(idx)  # Re-initialize the first frame

# Connect buttons to their functions
run_button.clicked.connect(run)
pause_button.clicked.connect(pause)
reset_button.clicked.connect(reset)

# Start a QTimer to call the update function periodically
timer = QtCore.QTimer()
timer.timeout.connect(update)
timer.start(100)  # Update every 100ms (10 frames per second)

# Show the main window
main_win.show()

# Start the Qt event loop
if __name__ == '__main__':
    if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
        QtWidgets.QApplication.instance().exec_()

"""










"""
2) build IM onsky - identify mode that corresponds to optical gain
for a pure scaling term in iinput aberations see how ratio of modes change with ratio of scaling factor. This should be constant besides the mode
most correlated with optical gain 
3) try get zonal closed loop working with proper DM registration and interpolation. Create basis based on registration. 
Then define tip / tilt and optical gain modes and filter them out 


I have a square deformable mirror that will actuator on a circular pupil of light. Only a few actuators in the circular region are registered with the actual pupil (lay within the illuminated region). I want to define a basis on these actuators, but also need the actuators outside the illuminated pupil (outside the registered basis) to vary smoothly with the registered actuators due to inter-actuator coupling. """



# Call the function to add the project root

import numpy as np
import matplotlib.pyplot as plt
from types import SimpleNamespace
import importlib 
from matplotlib.gridspec import GridSpec
from mpl_toolkits.axes_grid1 import make_axes_locatable
import os 
import sys
#import aotools

def add_project_root_to_sys_path(project_root_name="BaldrApp"):
    """
    Adds the project root directory to sys.path to allow importing from shared modules.
    
    Args:
        project_root_name (str): The name of the project root directory.
    """
    try:
        # Attempt to use __file__ to get the current script's directory
        current_dir = os.path.dirname(os.path.abspath(__file__))
    except NameError:
        # Fallback to current working directory (useful in interactive environments)
        current_dir = os.getcwd()

    # Traverse up the directory tree to find the project root
    project_root = current_dir
    while True:
        if os.path.basename(project_root) == project_root_name:
            break
        new_project_root = os.path.dirname(project_root)
        if new_project_root == project_root:
            # Reached the filesystem root without finding the project root
            project_root = None
            break
        project_root = new_project_root

    if project_root and project_root not in sys.path:
        sys.path.append(project_root)
        print(f"Added '{project_root}' to sys.path")
    elif project_root is None:
        print(f"Error: '{project_root_name}' directory not found in the directory hierarchy.")
    else:
        print(f"'{project_root}' is already in sys.path")

# Call the function to add the project root
add_project_root_to_sys_path()
from baldrapp.common import baldr_core as bldr
from baldrapp.common import DM_basis as gen_basis
from baldrapp.common import utilities as util
from baldrapp.common import DM_registration as DM_reg
from baldrapp.common import phasescreens

""" 
1) how square vs circular and different radii pupil region classification affects eigenvectors 

"""


################## TEST 0 
# configure our zwfs 
grid_dict = {
    "D":1, # diameter of beam 
    "N" : 64, # number of pixels across pupil diameter
    "padding_factor" : 4, # how many pupil diameters fit into grid x axis
    # TOTAL NUMBER OF PIXELS = padding_factor * N 
    }

optics_dict = {
    "wvl0" :1.65e-6, # central wavelength (m) 
    "F_number": 21.2, # F number on phasemask
    "mask_diam": 1.06, # diameter of phaseshifting region in diffraction limit units (physical unit is mask_diam * 1.22 * F_number * lambda)
    "theta": 1.57079, # phaseshift of phasemask 
}

dm_dict = {
    "dm_model":"BMC-multi-3.5",
    "actuator_coupling_factor":0.7,# std of in actuator spacing of gaussian IF applied to each actuator. (e.g actuator_coupling_factor = 1 implies std of poke is 1 actuator across.)
    "dm_pitch":1,
    "dm_aoi":0, # angle of incidence of light on DM 
    "opd_per_cmd" : 3e-6, # peak opd applied at center of actuator per command unit (normalized between 0-1) 
    "flat_rmse" : 20e-9 # std (m) of flatness across Flat DM  
    }

grid_ns = SimpleNamespace(**grid_dict)
optics_ns = SimpleNamespace(**optics_dict)
dm_ns = SimpleNamespace(**dm_dict)

detector = (4,4) 

zwfs_ns = bldr.init_zwfs(grid_ns, optics_ns, dm_ns)

# CALIBRATION FIELD PARAMETERS
cal_opd_input  = 0 * zwfs_ns.grid.pupil_mask *  np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

cal_opd_internal = 10e-9 * zwfs_ns.grid.pupil_mask * np.random.randn( *zwfs_ns.grid.pupil_mask.shape)

cal_amp_input = 1e4 * zwfs_ns.grid.pupil_mask



basis = 'Zonal_pinned_edges'
Nmodes = 100
M2C_0 = gen_basis.construct_command_basis( basis= basis, number_of_modes = Nmodes, without_piston=True).T  

zwfs_ns = bldr.classify_pupil_regions( cal_opd_input ,  cal_amp_input ,  cal_opd_internal,  zwfs_ns , detector=detector ) 

zwfs_ns = bldr.build_IM( zwfs_ns ,  calibration_opd_input = cal_opd_input  , calibration_amp_input = cal_amp_input , \
            opd_internal = cal_opd_internal,  basis = basis, Nmodes =  Nmodes, poke_amp = 0.05, poke_method = 'double_sided_poke',\
                imgs_to_mean = 1, detector=detector)



### 
# DM registration 
###

# get inner corners for estiamting DM center in pixel space (have to deal seperately with pinned actuator basis)
if zwfs_ns.reco.IM.shape[0] == 100: # outer actuators are pinned, 
    corner_indicies = DM_reg.get_inner_square_indices(outer_size=10, inner_offset=3, without_outer_corners=False)
    
elif zwfs_ns.reco.IM.shape[0] == 140: # outer acrtuators are free 
    print(140)
    corner_indicies = DM_reg.get_inner_square_indices(outer_size=12, inner_offset=4, without_outer_corners=True)
else:
    print("CASE NOT MATCHED  d['I2M'].data.shape = { d['I2M'].data.shape}")
    
img_4_corners = []
dm_4_corners = []
for i in corner_indicies:
    dm_4_corners.append( np.where( M2C_0[i] )[0][0] )
    #dm2px.get_DM_command_in_2D( d['M2C'].data[:,i]  # if you want to plot it 

    tmp = np.zeros( zwfs_ns.pupil_regions.pupil_filt.shape )
    tmp.reshape(-1)[zwfs_ns.pupil_regions.pupil_filt.reshape(-1)] = zwfs_ns.reco.IM[i] 

    #plt.imshow( tmp ); plt.show()
    img_4_corners.append( abs(tmp ) )

#plt.imshow( np.sum( tosee, axis=0 ) ); plt.show()

# dm_4_corners should be an array of length 4 corresponding to the actuator index in the (flattened) DM command space
# img_4_corners should be an array of length 4xNxM where NxM are the image dimensions.
# !!! It is very important that img_4_corners are registered in the same order as dm_4_corners !!!
transform_dict = DM_reg.calibrate_transform_between_DM_and_image( dm_4_corners, img_4_corners, debug=True, fig_path = None )


fig = plt.figure(3)
N0 = bldr.get_N0( cal_opd_input , cal_amp_input, cal_opd_internal, zwfs_ns  , detector=detector )
im = plt.imshow( N0 )
cbar = fig.colorbar(im, ax=plt.gca(), pad=0.01)
cbar.set_label(r'Intensity', fontsize=15, labelpad=10)

plt.scatter(transform_dict['actuator_coord_list_pixel_space'][:, 0],\
    transform_dict['actuator_coord_list_pixel_space'][:, 1], \
        color='blue', marker='.', label = 'DM actuators')

plt.legend() 
plt.show() 







### 
# Build control model  
#  apply Kolmogorov statistics to DM and measure intensity response, fit linear model to data 
#  then try fitting more advanced models, compare improvement 
###

i = bldr.get_frame( cal_opd_input , cal_amp_input, cal_opd_internal, zwfs_ns )
interpolated_intensities = DM_reg.interpolate_pixel_intensities(image = i, pixel_coords = transform_dict['actuator_coord_list_pixel_space'])

act_per_it = 0.5 # how many actuators does the screen pass per iteration 
V = 10 / act_per_it  / zwfs_ns.grid.D #m/s (10 actuators across pupil on DM)
#scrn = aotools.infinitephasescreen.PhaseScreenVonKarman(nx_size= int( 12 / act_per_it ) , pixel_scale= zwfs_ns.grid.D / zwfs_ns.grid.N , r0=0.1, L0=12)
scrn = phasescreens.PhaseScreenVonKarman(nx_size= int( 12 / act_per_it ) , pixel_scale= zwfs_ns.grid.D / zwfs_ns.grid.N , r0=0.1, L0=12)
corner_indicies = [0, 11, 11 * 12, -1] # DM corner indidices



I0_before = bldr.get_I0( cal_opd_input , cal_amp_input, cal_opd_internal, zwfs_ns , detector=detector )
N0_before = bldr.get_N0( cal_opd_input , cal_amp_input, cal_opd_internal, zwfs_ns  , detector=detector )

dm_I0_before = DM_reg.interpolate_pixel_intensities(image = I0_before, pixel_coords = transform_dict['actuator_coord_list_pixel_space'])
dm_N0_before = DM_reg.interpolate_pixel_intensities(image = N0_before, pixel_coords = transform_dict['actuator_coord_list_pixel_space'])
training_data = {'I0':I0_before, 'dm_I0':dm_I0_before , 'dm_N0':dm_N0_before , 'N0':N0_before, 'dm_flat':zwfs_ns.dm.dm_flat, 'rmse':[], 'dm_cmd':[], 'i':[], 'interp_i':[]} 

no_iters = 100 
for i in range(no_iters):
    print( i / no_iters )
    # roll phase screen
    scrn.add_row()
    # bin phase screen onto DM space 
    dm_scrn = util.create_phase_screen_cmd_for_DM(scrn,  scaling_factor=0.05, drop_indicies = [0, 11, 11 * 12, -1] , plot_cmd=False)
    # update DM command 
    zwfs_ns.dm.current_cmd = zwfs_ns.dm.dm_flat + dm_scrn
    # get ZWFS intensity 
    i = bldr.get_frame( cal_opd_input , cal_amp_input, cal_opd_internal, zwfs_ns , detector=detector )
    # interpolate onto registered DM actuators
    interpolated_intensities = DM_reg.interpolate_pixel_intensities(image = i , pixel_coords = transform_dict['actuator_coord_list_pixel_space'])

    training_data['dm_cmd'].append( dm_scrn )
    training_data['i'].append( i )
    training_data['rmse'].append( np.std( dm_scrn ) )
    training_data['interp_i'].append( interpolated_intensities )
 












i = -1
fig, ax = plt.subplots(1, 2, figsize=(10, 5))

im1 = ax[0].imshow(util.get_DM_command_in_2D(training_data['dm_cmd'][i]))
cbar1 = plt.colorbar(im1, ax=ax[0])
cbar1.ax.set_ylabel('Colorbar 1', rotation=270, labelpad=15)

im2 = ax[1].imshow(util.get_DM_command_in_2D(training_data['interp_i'][i] - dm_I0_before))
#im2 = ax[1].imshow(util.get_DM_command_in_2D(training_data['interp_i'][i] ))
cbar2 = plt.colorbar(im2, ax=ax[1])
cbar2.ax.set_ylabel('Colorbar 2', rotation=270, labelpad=15)

plt.tight_layout()

plt.show()

#a = 65
#plt.plot( training_data['dm_cmd'], training_data['interp_i'] ,'.') ; plt.show()



# Extract the input (dm_cmd) and output (interp_i) from training data
X = training_data['dm_cmd']  # Shape: (N, 140) where N is the number of training samples
y = training_data['interp_i']  # Shape: (N, 140) corresponding outputs


from sklearn.linear_model import Ridge
from sklearn.model_selection import train_test_split
import numpy as np

# Split data into training and testing sets
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

# Initialize and fit the Ridge regression model
linear_model = Ridge(alpha=1.0)  # Regularization parameter can be tuned
linear_model.fit(X_train, y_train)

# Make predictions
y_pred_linear = linear_model.predict(X_test)

# Evaluate performance (e.g., RMSE)
rmse_linear = np.sqrt(np.mean((y_pred_linear - y_test) ** 2))
print(f'Linear Model RMSE: {rmse_linear}')



import numpy as np 
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from mpl_toolkits.axes_grid1 import make_axes_locatable
import datetime 
from astropy import units as u
from matplotlib.widgets import Slider
import matplotlib.animation as animation
import math
from configparser import ConfigParser
from types import SimpleNamespace

def ini_to_namespace(ini_file):
    # convert ini file to python namespace
    
    # Create a ConfigParser object and read the INI file
    config = ConfigParser()
    config.read(ini_file)

    # Initialize an empty SimpleNamespace
    namespace = SimpleNamespace()

    # Iterate through sections and keys to populate the namespace
    for section in config.sections():
        section_namespace = SimpleNamespace()
        for key, value in config.items(section):
            setattr(section_namespace, key, value)
        
        # Set the section as an attribute of the main namespace
        setattr(namespace, section, section_namespace)

    return namespace


from configparser import ConfigParser
from types import SimpleNamespace

def ini_to_namespace(ini_file):
    # Create a ConfigParser object and read the INI file
    config = ConfigParser()
    config.read(ini_file)

    # Initialize an empty SimpleNamespace
    namespace = SimpleNamespace()

    # Iterate through sections and keys to populate the namespace
    for section in config.sections():
        section_namespace = SimpleNamespace()
        for key, value in config.items(section):
            # Try automatic type conversion using ConfigParser methods
            if config.has_option(section, key):
                # First attempt to convert to an integer
                try:
                    converted_value = config.getint(section, key)
                except ValueError:
                    # If it's not an int, try converting to a float
                    try:
                        converted_value = config.getfloat(section, key)
                    except ValueError:
                        # If it's not a float, check if it's a boolean
                        try:
                            converted_value = config.getboolean(section, key)
                        except ValueError:
                            # Fallback to original string if no conversion works
                            converted_value = value
            
            setattr(section_namespace, key, converted_value)
        
        # Set the section as an attribute of the main namespace
        setattr(namespace, section, section_namespace)

    return namespace



def get_DM_command_in_2D(cmd,Nx_act=12):
    # function so we can easily plot the DM shape (since DM grid is not perfectly square raw cmds can not be plotted in 2D immediately )
    #puts nan values in cmd positions that don't correspond to actuator on a square grid until cmd length is square number (12x12 for BMC multi-2.5 DM) so can be reshaped to 2D array to see what the command looks like on the DM.
    corner_indices = [0, Nx_act-1, Nx_act * (Nx_act-1), Nx_act*Nx_act]
    cmd_in_2D = list(cmd.copy())
    for i in corner_indices:
        cmd_in_2D.insert(i,np.nan)
    return( np.array(cmd_in_2D).reshape(Nx_act,Nx_act) )


def insert_concentric(smaller_array, larger_array):
    # Get the shapes of both arrays
    N, M = smaller_array.shape
    P, Q = larger_array.shape

    # Check if the smaller array can fit in the larger array
    if N > P or M > Q:
        raise ValueError("Smaller array must have dimensions less than or equal to the larger array.")

    # Find the starting indices to center the smaller array in the larger array
    start_row = (P - N) // 2
    start_col = (Q - M) // 2

    # Create a copy of the larger array to avoid modifying the input directly
    result_array = larger_array.copy()

    # Insert the smaller array into the larger array
    result_array[start_row:start_row + N, start_col:start_col + M] = smaller_array

    return result_array



def crop_pupil(pupil, image):
    """
    Detects the boundary of a pupil in a binary mask (with pupil = 1 and background = 0)
    and crops both the pupil mask and the corresponding image to contain just the pupil.
    
    Parameters:
    - pupil: A 2D NumPy array (binary) representing the pupil (1 inside the pupil, 0 outside).
    - image: A 2D NumPy array of the same shape as 'pupil' representing the image to be cropped.
    
    Returns:
    - cropped_pupil: The cropped pupil mask.
    - cropped_image: The cropped image based on the pupil's bounding box.
    """
    # Ensure both arrays have the same shape
    if pupil.shape != image.shape:
        raise ValueError("Pupil and image must have the same dimensions.")

    # Sum along the rows (axis=1) to find the non-zero rows (pupil region)
    row_sums = np.sum(pupil, axis=1)
    non_zero_rows = np.where(row_sums > 0)[0]

    # Sum along the columns (axis=0) to find the non-zero columns (pupil region)
    col_sums = np.sum(pupil, axis=0)
    non_zero_cols = np.where(col_sums > 0)[0]

    # Get the bounding box of the pupil by identifying the min and max indices
    row_start, row_end = non_zero_rows[0], non_zero_rows[-1] + 1
    col_start, col_end = non_zero_cols[0], non_zero_cols[-1] + 1

    # Crop both the pupil and the image
    cropped_pupil = pupil[row_start:row_end, col_start:col_end]
    cropped_image = image[row_start:row_end, col_start:col_end]

    return cropped_pupil, cropped_image






def create_phase_screen_cmd_for_DM(scrn,  scaling_factor=0.1, drop_indicies = None, plot_cmd=False):
    """
    aggregate a scrn (aotools.infinitephasescreen object) onto a DM command space. phase screen is normalized by
    between +-0.5 and then scaled by scaling_factor. Final DM command values should
    always be between -0.5,0.5 (this should be added to a flat reference so flat reference + phase screen should always be bounded between 0-1). phase screens are usually a NxN matrix, while DM is MxM with some missing pixels (e.g. 
    corners). drop_indicies is a list of indicies in the flat MxM DM array that should not be included in the command space. 
    """

    #print('----------\ncheck phase screen size is multiple of DM\n--------')
    
    Nx_act = 12 #number of actuators across DM diameter
    
    scrn_array = ( scrn.scrn - np.min(scrn.scrn) ) / (np.max(scrn.scrn) - np.min(scrn.scrn)) - 0.5 # normalize phase screen between -0.5 - 0.5 
    
    size_factor = int(scrn_array.shape[0] / Nx_act) # how much bigger phase screen is to DM shape in x axis. Note this should be an integer!!
    
    # reshape screen so that axis 1,3 correspond to values that should be aggregated 
    scrn_to_aggregate = scrn_array.reshape(scrn_array.shape[0]//size_factor, size_factor, scrn_array.shape[1]//size_factor, size_factor)
    
    # now aggreagate and apply the scaling factor 
    scrn_on_DM = scaling_factor * np.mean( scrn_to_aggregate, axis=(1,3) ).reshape(-1) 

    #If DM is missing corners etc we set these to nan and drop them before sending the DM command vector
    #dm_cmd =  scrn_on_DM.to_list()
    if drop_indicies is not None:
        for i in drop_indicies:
            scrn_on_DM[i]=np.nan
             
    if plot_cmd: #can be used as a check that the command looks right!
        fig,ax = plt.subplots(1,2,figsize=(12,6))
        im0 = ax[0].imshow( scrn_on_DM.reshape([Nx_act,Nx_act]) )
        ax[0].set_title('DM command (averaging offset)')
        im1 = ax[1].imshow(scrn.scrn)
        ax[1].set_title('original phase screen')
        plt.colorbar(im0, ax=ax[0])
        plt.colorbar(im1, ax=ax[1]) 
        plt.show() 

    dm_cmd =  list( scrn_on_DM[np.isfinite(scrn_on_DM)] ) #drop non-finite values which should be nan values created from drop_indicies array
    return(dm_cmd) 





def magnitude_to_photon_flux(magnitude, band, wavelength):
    """
    Convert stellar magnitude in a given band to photon flux (photons / s / m^2 / nm).
    
    ***EXPERIMENTAL  - need to verify results 
    
    Parameters:
    - magnitude: The magnitude of the star.
    - band: The name of the filter (e.g., 'V', 'J', 'H').
    - wavelength: The central wavelength of the filter in nm.
    
    Returns:
    - photon_flux: The number of photons per second per square meter per nanometer.
    """

    from astropy.constants import h, c
    # Zero points in energy flux for different bands (in erg/s/cm^2/Å)
    zero_point_flux = {
        'V': 3.63e-9 * u.erg / (u.cm**2 * u.s * u.AA),  # V-band zero point
        'J': 3.13e-10 * u.erg / (u.cm**2 * u.s * u.AA), # J-band zero point
        'H': 1.16e-10 * u.erg / (u.cm**2 * u.s * u.AA), # H-band zero point
        # Add more bands as needed
    }
    
    if band not in zero_point_flux:
        raise ValueError(f"Unknown band: {band}. Available bands are {list(zero_point_flux.keys())}")
    
    # Convert magnitude to energy flux density (f_lambda in erg/s/cm^2/Å)
    f_lambda = zero_point_flux[band] * 10**(-0.4 * magnitude)
    
    # Convert wavelength to meters
    wavelength_m = (wavelength * u.nm).to(u.m)
    
    # Convert energy flux density to W/m^2/nm
    f_lambda_si = f_lambda.to(u.W / (u.m**2 * u.nm), equivalencies=u.spectral_density(wavelength_m))
    
    # Calculate the energy per photon (in joules) at the given wavelength
    energy_per_photon = (h * c / wavelength_m).to(u.J)  # Energy per photon at this wavelength
    
    # Calculate photon flux (photons/s/m^2/nm)
    photon_flux = f_lambda_si / energy_per_photon.value  # Explicitly divide by the scalar value of energy_per_photon
    
    # Return photon flux in the appropriate units (photon/s/m^2/nm)
    return photon_flux.value





###### MODELLING MIRROR SCRATCHES


def apply_parabolic_scratches(array, dx, dy, list_a, list_b, list_c, width_list, depth_list):
    """
    Apply multiple parabolic scratches to a 2D array based on input parameters.

    Parameters:
    array (2D numpy array): The input 2D array to which the scratches will be applied.
    dx (float): Pixel scale in the x direction.
    dy (float): Pixel scale in the y direction.
    list_a, list_b, list_c (lists of floats): Lists of a, b, c coefficients for each parabola (y = a*x^2 + b*x + c).
    width_list (list of floats): List of widths for each scratch around the parabolic contour.
    depth_list (list of floats): List of depths for each scratch.

    Returns:
    Modified 2D numpy array with the parabolic scratches applied.
    """
    num_pixels_y, num_pixels_x = array.shape

    # Generate x and y coordinates corresponding to pixel locations
    x_vals = np.linspace(-num_pixels_x/2 * dx, num_pixels_x/2 * dx, num_pixels_x)
    y_vals = np.linspace(-num_pixels_y/2 * dy, num_pixels_y/2 * dy, num_pixels_y)
    X, Y = np.meshgrid(x_vals, y_vals)
    
    # Apply each parabolic scratch
    for a, b, c, width, depth in zip(list_a, list_b, list_c, width_list, depth_list):
        # Compute the parabolic contour y = a*x^2 + b*x + c for each scratch
        parabolic_curve_y = a * X**2 + b * X + c
        
        # Apply the scratch around the parabolic curve with constant depth and width
        for i in range(num_pixels_y):
            for j in range(num_pixels_x):
                # Compute the distance to the parabolic contour
                distance_to_parabola = np.abs(Y[i, j] - parabolic_curve_y[i, j])
                
                # If the point is within the width of the scratch, modify the array
                if distance_to_parabola <= width / 2:
                    array[i, j] -= depth

    return array

# # Example usage
# num_pixels_x, num_pixels_y = 100, 100  # Size of the array (100x100)
# dx, dy = 0.1, 0.1  # Pixel scale in the x and y directions
# input_array = np.full((num_pixels_y, num_pixels_x), 10)  # Constant background array

# # Lists of parabolic parameters, widths, and depths for the scratches
# list_a = [0.5, 0.7]
# list_b = [0, 0]
# list_c = [0, 2]
# width_list = [0.5, 0.2]  # Width of the scratches
# depth_list = [2, 3]  # Depth of the scratches

# # Apply the scratches
# modified_array = apply_parabolic_scratches(input_array, dx, dy, list_a, list_b, list_c, width_list, depth_list)

# # Visualize the result
# import matplotlib.pyplot as plt
# plt.imshow(modified_array, cmap='hot', extent=[-num_pixels_x/2*dx, num_pixels_x/2*dx, -num_pixels_y/2*dy, num_pixels_y/2*dy])
# plt.colorbar(label='Value')
# plt.title('2D Array with Multiple Parabolic Scratches')
# plt.show()





#### PLOTTING 


def nice_heatmap_subplots( im_list , xlabel_list=None, ylabel_list=None, title_list=None, cbar_label_list=None, fontsize=15, cbar_orientation = 'bottom', axis_off=True, vlims=None, savefig=None):

    n = len(im_list)
    fs = fontsize
    fig = plt.figure(figsize=(5*n, 5))

    for a in range(n) :
        ax1 = fig.add_subplot(int(f'1{n}{a+1}'))

        if vlims is not None:
            im1 = ax1.imshow(  im_list[a] , vmin = vlims[a][0], vmax = vlims[a][1])
        else:
            im1 = ax1.imshow(  im_list[a] )
        if title_list is not None:
            ax1.set_title( title_list[a] ,fontsize=fs)
        if xlabel_list is not None:
            ax1.set_xlabel( xlabel_list[a] ,fontsize=fs) 
        if ylabel_list is not None:
            ax1.set_ylabel( ylabel_list[a] ,fontsize=fs) 
        ax1.tick_params( labelsize=fs ) 

        if axis_off:
            ax1.axis('off')
        divider = make_axes_locatable(ax1)
        if cbar_orientation == 'bottom':
            cax = divider.append_axes('bottom', size='5%', pad=0.05)
            cbar = fig.colorbar( im1, cax=cax, orientation='horizontal')
                
        elif cbar_orientation == 'top':
            cax = divider.append_axes('top', size='5%', pad=0.05)
            cbar = fig.colorbar( im1, cax=cax, orientation='horizontal')
                
        else: # we put it on the right 
            cax = divider.append_axes('right', size='5%', pad=0.05)
            cbar = fig.colorbar( im1, cax=cax, orientation='vertical')  
        
        if cbar_label_list is not None:
            cbar.set_label( cbar_label_list[a], rotation=0,fontsize=fs)
        cbar.ax.tick_params(labelsize=fs)
    if savefig is not None:
        plt.savefig( savefig , bbox_inches='tight', dpi=300) 

    #plt.show()
    
    
     

def nice_DM_plot( data, savefig=None ): #for a 140 actuator BMC 3.5 DM
    fig,ax = plt.subplots(1,1)
    if len( np.array(data).shape ) == 1: 
        ax.imshow( get_DM_command_in_2D(data) )
    else: 
        ax.imshow( data )
    #ax.set_title('poorly registered actuators')
    ax.grid(True, which='minor',axis='both', linestyle='-', color='k', lw=2 )
    ax.set_xticks( np.arange(12) - 0.5 , minor=True)
    ax.set_yticks( np.arange(12) - 0.5 , minor=True)
    if savefig is not None:
        plt.savefig( savefig , bbox_inches='tight', dpi=300) 



def plot_data_and_residuals(x, y_meas, y_model, xlabel, ylabel, residual_ylabel, label_1=None, label_2=None, savefig=None):
    # Calculate residuals
    residuals = y_meas - y_model

    # Create a figure with two subplots: one for the data and one for the residuals
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(8, 6), gridspec_kw={'height_ratios': [3, 1]}, sharex=True)
    
    # First subplot: measured and modeled data
    if label_1 is None:
        ax1.plot(x, y_meas, '-', label='Measured Data', color='blue', markersize=2, alpha =0.3)
    else: 
        ax1.plot(x, y_meas, '-', label=label_1, color='blue', markersize=2, alpha =0.3)
    if label_2 is None:    
        ax1.plot(x, y_model, '.', label='Modeled Data', color='red', linewidth=2, alpha =0.3)
    else:
        ax1.plot(x, y_model, '.', label=label_2, color='red', linewidth=2, alpha =0.3)
    #ax1.set_xlabel(xlabel)
    ax1.set_ylabel(ylabel)
    ax1.legend()
    ax1.grid(True)

    # Second subplot: residuals
    ax2.plot(x, residuals, '.', color='green', markersize=5, alpha =0.3)
    ax2.axhline(0, color='black', linewidth=1, linestyle='--')  # Horizontal line at zero
    ax2.set_xlabel(xlabel)
    ax2.set_ylabel(residual_ylabel)
    ax2.grid(True)

    # Adjust layout for better spacing between subplots
    plt.tight_layout()
    
    if savefig is not None:
        plt.savefig(savefig, bbox_inches='tight', dpi=300  )
    # Show the plot
    plt.show()
    
    

def create_telem_mosaic(image_list, image_title_list, image_colorbar_list, 
                  plot_list, plot_title_list, plot_xlabel_list, plot_ylabel_list):
    """
    Creates a 3-row mosaic layout with:
    - First row: images with colorbars below
    - Second and third rows: plots with titles and axis labels
    
    Parameters:
    - image_list: List of image data for the first row (4 images)
    - image_title_list: List of titles for the first row images
    - image_colorbar_list: List of colorbars (True/False) for each image in the first row
    - plot_list: List of plot data for second and third rows (4 plots, 2 per row)
    - plot_title_list: List of titles for each plot
    - plot_xlabel_list: List of x-axis labels for each plot
    - plot_ylabel_list: List of y-axis labels for each plot
    """
    
    # Create a figure with constrained layout and extra padding
    fig = plt.figure(constrained_layout=True, figsize=(10, 8))
    plt.subplots_adjust(wspace=0.4, hspace=0.4)
    
    # Create GridSpec with 3 rows and different numbers of columns
    gs = GridSpec(3, 4, figure=fig, height_ratios=[1, 1, 1])
    
    # Top row: 4 columns with colorbars
    for i in range(4):
        ax = fig.add_subplot(gs[0, i])
        img = image_list[i]
        im = ax.imshow(img, cmap='viridis')  # Modify colormap if needed
        ax.set_title(image_title_list[i])
        
        # Optionally add a colorbar below the image
        if image_colorbar_list[i]:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes('bottom', size='5%', pad=0.2)
            fig.colorbar(im, cax=cax, orientation='horizontal')
    
    # Middle row: 2 columns, each spanning 2 grid columns
    for i in range(2):
        ax = fig.add_subplot(gs[1, 2*i:2*i+2])
        data = plot_list[i]
        ax.plot(data)
        ax.set_title(plot_title_list[i])
        ax.set_xlabel(plot_xlabel_list[i])
        ax.set_ylabel(plot_ylabel_list[i])

    # Bottom row: 2 columns, each spanning 2 grid columns
    for i in range(2, 4):
        ax = fig.add_subplot(gs[2, 2*(i-2):2*(i-2)+2])
        data = plot_list[i]
        
        ax.plot(data)
        ax.set_title(plot_title_list[i])
        ax.set_xlabel(plot_xlabel_list[i])
        ax.set_ylabel(plot_ylabel_list[i])
    
    # Show the plot
    plt.show()



def plot_eigenmodes( zwfs_ns , save_path = None ):
    
    tstamp = datetime.datetime.now().strftime("%d-%m-%YT%H.%M.%S")

    U,S,Vt = np.linalg.svd( zwfs_ns.reco.IM, full_matrices=True)

    #singular values
    plt.figure() 
    plt.semilogy(S) #/np.max(S))
    #plt.axvline( np.pi * (10/2)**2, color='k', ls=':', label='number of actuators in pupil')
    plt.legend() 
    plt.xlabel('mode index')
    plt.ylabel('singular values')

    if save_path is not None:
        plt.savefig(save_path +  f'singularvalues_{tstamp}.png', bbox_inches='tight', dpi=200)
    plt.show()
    
    # THE IMAGE MODES 
    n_row = round( np.sqrt( zwfs_ns.reco.M2C_0.shape[0]) ) - 1
    fig,ax = plt.subplots(n_row  ,n_row ,figsize=(30,30))
    plt.subplots_adjust(hspace=0.1,wspace=0.1)
    for i,axx in enumerate(ax.reshape(-1)):
        # we filtered circle on grid, so need to put back in grid
        tmp =  zwfs_ns.pupil_regions.pupil_filt.copy()
        vtgrid = np.zeros(tmp.shape)
        vtgrid[tmp] = Vt[i]
        r1,r2,c1,c2 = 10,-10,10,-10
        axx.imshow( vtgrid.reshape(zwfs_ns.reco.I0.shape )[r1:r2,c1:c2] ) #cp_x2-cp_x1,cp_y2-cp_y1) )
        #axx.set_title(f'\n\n\nmode {i}, S={round(S[i]/np.max(S),3)}',fontsize=5)
        #
        axx.text( 10,10, f'{i}',color='w',fontsize=4)
        axx.text( 10,20, f'S={round(S[i]/np.max(S),3)}',color='w',fontsize=4)
        axx.axis('off')
        #plt.legend(ax=axx)
    plt.tight_layout()

    if save_path is not None:
        plt.savefig(save_path + f'det_eignmodes_{tstamp}.png',bbox_inches='tight',dpi=200)
    plt.show()
    
    # THE DM MODES 

    # NOTE: if not zonal (modal) i might need M2C to get this to dm space 
    # if zonal M2C is just identity matrix. 
    fig,ax = plt.subplots(n_row, n_row, figsize=(30,30))
    plt.subplots_adjust(hspace=0.1,wspace=0.1)
    for i,axx in enumerate(ax.reshape(-1)):
        axx.imshow( get_DM_command_in_2D( zwfs_ns.reco.M2C_0.T @ U.T[i] ) )
        #axx.set_title(f'mode {i}, S={round(S[i]/np.max(S),3)}')
        axx.text( 1,2,f'{i}',color='w',fontsize=6)
        axx.text( 1,3,f'S={round(S[i]/np.max(S),3)}',color='w',fontsize=6)
        axx.axis('off')
        #plt.legend(ax=axx)
    plt.tight_layout()
    if save_path is not None:
        plt.savefig(save_path +  f'dm_eignmodes_{tstamp}.png',bbox_inches='tight',dpi=200)
    plt.show()




def display_images_with_slider(image_lists, plot_titles=None, cbar_labels=None):
    """
    Displays multiple images or 1D plots from a list of lists with a slider to control the shared index.
    
    Parameters:
    - image_lists: list of lists where each inner list contains either 2D arrays (images) or 1D arrays (scalars).
                   The inner lists must all have the same length.
    - plot_titles: list of strings, one for each subplot. Default is None (no titles).
    - cbar_labels: list of strings, one for each colorbar. Default is None (no labels).
    """
    
    # Check that all inner lists have the same length
    assert all(len(lst) == len(image_lists[0]) for lst in image_lists), "All inner lists must have the same length."
    
    # Number of rows and columns based on the number of plots
    num_plots = len(image_lists)
    ncols = math.ceil(math.sqrt(num_plots))  # Number of columns for grid
    nrows = math.ceil(num_plots / ncols)     # Number of rows for grid
    
    num_frames = len(image_lists[0])

    # Create figure and axes
    fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(5 * ncols, 5 * nrows))
    plt.subplots_adjust(bottom=0.2)

    # Flatten axes array for easier iteration
    axes = axes.flatten() if num_plots > 1 else [axes]

    # Store the display objects for each plot (either imshow or line plot)
    img_displays = []
    line_displays = []
    
    # Get max/min values for 1D arrays to set static axis limits
    max_values = [max(lst) if not isinstance(lst[0], np.ndarray) else None for lst in image_lists]
    min_values = [min(lst) if not isinstance(lst[0], np.ndarray) else None for lst in image_lists]

    for i, ax in enumerate(axes[:num_plots]):  # Only iterate over the number of plots
        # Check if the first item in the list is a 2D array (an image) or a scalar
        if isinstance(image_lists[i][0], np.ndarray) and image_lists[i][0].ndim == 2:
            # Use imshow for 2D data (images)
            img_display = ax.imshow(image_lists[i][0], cmap='viridis')
            img_displays.append(img_display)
            line_displays.append(None)  # Placeholder for line plots
            
            # Add colorbar if it's an image
            cbar = fig.colorbar(img_display, ax=ax)
            if cbar_labels and i < len(cbar_labels) and cbar_labels[i] is not None:
                cbar.set_label(cbar_labels[i])

        else:
            # Plot the list of scalar values up to the initial index
            line_display, = ax.plot(np.arange(len(image_lists[i])), image_lists[i], color='b')
            line_display.set_data(np.arange(1), image_lists[i][:1])  # Start with only the first value
            ax.set_xlim(0, len(image_lists[i]))  # Set x-axis to full length of the data
            ax.set_ylim(min_values[i], max_values[i])  # Set y-axis to cover the full range
            line_displays.append(line_display)
            img_displays.append(None)  # Placeholder for image plots

        # Set plot title if provided
        if plot_titles and i < len(plot_titles) and plot_titles[i] is not None:
            ax.set_title(plot_titles[i])

    # Remove any unused axes
    for ax in axes[num_plots:]:
        ax.remove()

    # Slider for selecting the frame index
    ax_slider = plt.axes([0.2, 0.05, 0.65, 0.03], facecolor='lightgoldenrodyellow')
    frame_slider = Slider(ax_slider, 'Frame', 0, num_frames - 1, valinit=0, valstep=1)

    # Update function for the slider
    def update(val):
        index = int(frame_slider.val)  # Get the selected index from the slider
        for i, (img_display, line_display) in enumerate(zip(img_displays, line_displays)):
            if img_display is not None:
                # Update the image data for 2D data
                img_display.set_data(image_lists[i][index])
            if line_display is not None:
                # Update the line plot for scalar values (plot up to the selected index)
                line_display.set_data(np.arange(index), image_lists[i][:index])
        fig.canvas.draw_idle()  # Redraw the figure

    # Connect the slider to the update function
    frame_slider.on_changed(update)

    plt.show()



def display_images_as_movie(image_lists, plot_titles=None, cbar_labels=None, save_path="output_movie.mp4", fps=5):
    """
    Creates an animation from multiple images or 1D plots from a list of lists and saves it as a movie.
    
    Parameters:
    - image_lists: list of lists where each inner list contains either 2D arrays (images) or 1D arrays (scalars).
                   The inner lists must all have the same length.
    - plot_titles: list of strings, one for each subplot. Default is None (no titles).
    - cbar_labels: list of strings, one for each colorbar. Default is None (no labels).
    - save_path: path where the output movie will be saved.
    - fps: frames per second for the output movie.
    """
    
    # Check that all inner lists have the same length
    assert all(len(lst) == len(image_lists[0]) for lst in image_lists), "All inner lists must have the same length."
    
    # Number of rows and columns based on the number of plots
    num_plots = len(image_lists)
    ncols = math.ceil(math.sqrt(num_plots))  # Number of columns for grid
    nrows = math.ceil(num_plots / ncols)     # Number of rows for grid
    
    num_frames = len(image_lists[0])

    # Create figure and axes
    fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(6 * ncols, 6 * nrows))
    plt.subplots_adjust(bottom=0.2)

    # Flatten axes array for easier iteration
    axes = axes.flatten() if num_plots > 1 else [axes]

    # Store the display objects for each plot (either imshow or line plot)
    img_displays = []
    line_displays = []
    
    # Get max/min values for 1D arrays to set static axis limits
    max_values = [max(lst) if not isinstance(lst[0], np.ndarray) else None for lst in image_lists]
    min_values = [min(lst) if not isinstance(lst[0], np.ndarray) else None for lst in image_lists]

    for i, ax in enumerate(axes[:num_plots]):  # Only iterate over the number of plots
        # Check if the first item in the list is a 2D array (an image) or a scalar
        if isinstance(image_lists[i][0], np.ndarray) and image_lists[i][0].ndim == 2:
            # Use imshow for 2D data (images)
            img_display = ax.imshow(image_lists[i][0], cmap='viridis')
            img_displays.append(img_display)
            line_displays.append(None)  # Placeholder for line plots
            
            # Add colorbar if it's an image
            cbar = fig.colorbar(img_display, ax=ax)
            if cbar_labels and i < len(cbar_labels) and cbar_labels[i] is not None:
                cbar.set_label(cbar_labels[i])

        else:
            # Plot the list of scalar values up to the initial index
            line_display, = ax.plot(np.arange(len(image_lists[i])), image_lists[i], color='b')
            line_display.set_data(np.arange(1), image_lists[i][:1])  # Start with only the first value
            ax.set_xlim(0, len(image_lists[i]))  # Set x-axis to full length of the data
            ax.set_ylim(min_values[i], max_values[i])  # Set y-axis to cover the full range
            line_displays.append(line_display)
            img_displays.append(None)  # Placeholder for image plots

        # Set plot title if provided
        if plot_titles and i < len(plot_titles) and plot_titles[i] is not None:
            ax.set_title(plot_titles[i])

    # Remove any unused axes
    for ax in axes[num_plots:]:
        ax.remove()

    # Function to update the frames
    def update_frame(frame_idx):
        for i, (img_display, line_display) in enumerate(zip(img_displays, line_displays)):
            if img_display is not None:
                # Update the image data for 2D data
                img_display.set_data(image_lists[i][frame_idx])
            if line_display is not None:
                # Update the line plot for scalar values (plot up to the current index)
                line_display.set_data(np.arange(frame_idx), image_lists[i][:frame_idx])
        return img_displays + line_displays

    # Create the animation
    ani = animation.FuncAnimation(fig, update_frame, frames=num_frames, blit=False, repeat=False)

    # Save the animation as a movie file
    ani.save(save_path, fps=fps, writer='ffmpeg')

    plt.show()

